@@ -26,6 +26,7 @@ import random
 import sys
 from itertools import chain
 from typing import (
+    Iterator,
     List,
     Mapping,
     Optional,
@@ -38,6 +39,7 @@ from py2048 import (
     SquareGameGrid,
     Directions,
     ExpectationError,
+    Line,
     NegativeIntegerError,
     Point,
     Snapshot,
@@ -51,17 +53,25 @@ logger = logging.getLogger(__name__)
 
 class Grid(SquareGameGrid):
     CELLCLASS: Type = Cell
-    # how many cells start non-zero
+    # how many Cells start non-zero
     STARTING_AMOUNT = 2
-    # what values can be seeded in each cell; tuple instead of set because
+    # what values can be seeded in each Cell; tuple instead of set because
     # random.choice doesn't work with sets
     SEEDING_VALUES = (2, 4)
     # the `is2048like` method will use this list to validate a goal
     # 2**11 == 2048
     NUMBERS: List[int] = [2 ** power for power in range(1, 12)]
+    # _AUTO_NUMBERS is used to autofill the grid (for testing and debugging)
+    _AUTO_NUMBERS = NUMBERS[:-1]
     # no power of 2 higher than CEILING will be accepted as the game's goal
     # sys.maxsize == (2**63)-1 on 64bits machines
     CEILING = sys.maxsize
+    SHIFTS = {
+        Directions.LEFT: (-1, 0),
+        Directions.RIGHT: (1, 0),
+        Directions.DOWN: (0, 1),
+        Directions.UP: (0, -1),
+    }
 
     cells = SquareGameGrid.values
 
@@ -83,7 +93,7 @@ class Grid(SquareGameGrid):
             cls.NUMBERS.append(highest)
         return number in cls.NUMBERS
 
-    def __init__(self, side: int = 4) -> None:
+    def __init__(self, side: int = 4, storing_snapshot: bool = True) -> None:
         logger.info("Creating a %dx%d Grid instance", side, side)
         if self.STARTING_AMOUNT < 1:
             raise Base2048Error(
@@ -96,11 +106,11 @@ class Grid(SquareGameGrid):
                 f"{side_squared} STARTING_AMOUNT's"
             )
         super().__init__(side)
-        self.vectors = {
-            Directions.LEFT: self.columns(),
-            Directions.RIGHT: self.columns(reverse=True),
-            Directions.UP: self.rows(),
-            Directions.DOWN: self.rows(reverse=True),
+        self._vectors_getters = {
+            Directions.LEFT: (self.columns, False),
+            Directions.RIGHT: (self.columns, True),
+            Directions.UP: (self.rows, False),
+            Directions.DOWN: (self.rows, True),
         }
         # the grid starts empty, so every Cell starts in the `empty_cells` set
         self.empty_cells: Set[Cell] = set(self.cells())
@@ -116,7 +126,7 @@ class Grid(SquareGameGrid):
         # the grid could start jammed if self.STARTING_AMOUNT == side ** 2
         if self.is_jammed:
             self.autofill()  # will also store a snapshot
-        else:
+        elif storing_snapshot:
             self.store_snapshot()
         self.check_integrity()
 
@@ -135,17 +145,20 @@ class Grid(SquareGameGrid):
         self.history.clear()
         self.check_integrity()
 
-    def store_snapshot(self) -> None:
-        """Store the current game state in `history`.
+    def store_snapshot(self, snapshot: Optional[Snapshot] = None) -> None:
+        """Store a game state in `history`.
+
+        If `snapshot` isn't provided, it defaults to the current game state.
         """
 
-        snapshot = {point: cell.number for point, cell in self.items()}
+        if snapshot is None:
+            snapshot = {point: cell.number for point, cell in self.items()}
         self.history.append(snapshot)
 
     def update_with_snapshot(self, snapshot: Snapshot) -> None:
         """Replace each Cell number with the corresponding `snapshot` value.
 
-        Also unlock every Cell.
+        Also unlock every Cell. This DOESN'T store the snapshot.
         """
 
         for point, number in snapshot.items():
@@ -164,8 +177,9 @@ class Grid(SquareGameGrid):
         if side != sqrt:
             name = classname(cls)
             raise ValueError(f"Cannot create {name} from a non-square mapping")
-        new = cls(side)
+        new = cls(side, storing_snapshot=False)
         new.update_with_snapshot(snapshot)
+        new.store_snapshot()
         return new
 
     def undo(self, ignore_empty: bool = True) -> None:
@@ -176,18 +190,20 @@ class Grid(SquareGameGrid):
             current state, one for the previous one
         """
 
-        try:
+        if len(self.history) >= 2:
             # the last snapshot is the current state, so we forget it
             del self.history[-1]
             # now the last snapshot is the previous state, so we retrieve and
             # forget it
-            snap = self.history.pop()
-        except IndexError:
+            past = self.history.pop()
+        else:
             if not ignore_empty:
-                raise
+                raise IndexError(
+                    "cannot undo last movement: too few snapshots available"
+                )
             return
-        self.update_with_snapshot(snap)
-        self.store_snapshot()
+        self.update_with_snapshot(past)
+        self.store_snapshot(past)
 
     @property
     def largest(self) -> int:
@@ -198,8 +214,9 @@ class Grid(SquareGameGrid):
         """Return whether every Cell is empty (ie, 0-numbered).
         """
 
-        assert len(self.empty_cells) <= len(self)
-        return len(self.empty_cells) == len(self)
+        empties, cells = len(self.empty_cells), len(self)
+        assert empties <= cells
+        return empties == cells
 
     def set_cell(self, cell: Cell, number: int) -> None:
         """Assign `number` to `Cell` and update `empty_cells`.
@@ -259,15 +276,7 @@ class Grid(SquareGameGrid):
         of the board, for example).
         """
 
-        x, y = cell.point
-        if to == Directions.LEFT:
-            x -= 1
-        elif to == Directions.RIGHT:
-            x += 1
-        elif to == Directions.DOWN:
-            y += 1
-        elif to == Directions.UP:
-            y -= 1
+        x, y = cell.point + self.SHIFTS[to]
         try:
             next_point = Point(x, y)  # NegativeIntegerError
             next_cell = self[next_point]  # KeyError
@@ -330,20 +339,12 @@ class Grid(SquareGameGrid):
         self.set_cell(pivot, new_number)
         return True
 
-    def drag(self, to: Directions) -> bool:
-        """Try to move every cell towards `to`, starting with the vectors
-        closest to the origin.
-
-        Increment the attempt counter, then try to move every cell in the
-        given direction, starting with the vectors closest to the origin.
-        If that changed anything, increment the cycle counter by 1 and seed
-        itself by 1.
-
-        :return: whether the game state changed
+    def _vectors_from_Directions(self, to: Directions) -> Iterator[Line]:
+        """Return each row or column, from closest to farthest from `to`.
         """
 
         try:
-            vectors = self.vectors[to]
+            method, reverse_boolean = self._vectors_getters[to]
         except KeyError:
             if isinstance(to, Directions):
                 # very bad
@@ -353,6 +354,21 @@ class Grid(SquareGameGrid):
             else:
                 # EVEN WORSE
                 raise ExpectationError(to, Directions)
+        return method(reverse=reverse_boolean)
+
+    def drag(self, to: Directions) -> bool:
+        """Try to move every Cell towards `to`, starting with the vectors
+        closest to the origin.
+
+        Increment the attempt counter, then try to move every Cell in the
+        given direction, starting with the vectors closest to the origin.
+        If that changed anything, increment the cycle counter by 1 and seed
+        itself by 1.
+
+        :return: whether the game state changed
+        """
+
+        vectors = self._vectors_from_Directions(to)
         self.attempt += 1
         logger.debug("Attempt increased to %d.", self.attempt)
         something_moved = False
@@ -396,16 +412,16 @@ class Grid(SquareGameGrid):
         return True
 
     def _autofill(self) -> None:
-        numbers = self.NUMBERS[:-1]
         for cell in self.cells():
             cell.unlock()
-            self.set_cell(cell, random.choice(numbers))
+            self.set_cell(cell, random.choice(self._AUTO_NUMBERS))
 
     def autofill(self, no_jamming: bool = True) -> None:
-        """Give each cell a random number.
+        """Give each Cell a random number.
 
         Store a snapshot afterwards. The numbers are always greater than 0 and
         smaller than 2048.
+        Also, unlock every Cell.
 
         :param bool no_jamming: if the result cannot be a jammed grid
         """
